#ifndef _DEVX_VERBS_H
#define _DEVX_VERBS_H

/* set up helper functions that we need to have a 'verbs' like
 * functionality over objects opened via DEVX
 */

#include <stddef.h>
#include <stdbool.h>
#include <unistd.h>

#include "config.h"
#include "nvmx_util.h"
#include "devx_compat.h"
#include <infiniband/mlx5dv.h>
#include "nvme_emu_log.h"
#include "utils.h"

#define DEVX_SQ_LOG  6
/* 2048 receive buffers in the queue */
//#define DEVX_RQ_LOG  11
#define DEVX_RQ_LOG  10

#define DEVX_CQ_LOG  14

#define DEVX_CQ_SIZE (1 << DEVX_CQ_LOG)
#define DEVX_SQ_SIZE (1 << DEVX_SQ_LOG)
#define DEVX_RQ_SIZE (1 << DEVX_RQ_LOG)

#define DEVX_CQ_POLL_TO 1600000
#define DEVX_CQ_POLL_INF (-1)

/* number of bits in the QP */
#define DEVX_IB_QPN_ORDER 24

/* mlx5_hw.h does not like mlx5dv.h. So define only what is needed */
enum {
    MLX5_HCA_CAP_OPMOD_GET_MAX  = 0,
    MLX5_HCA_CAP_OPMOD_GET_CUR  = 1,
};

enum {
    MLX5_CAP_GENERAL = 0,
    MLX5_CAP_ETHERNET_OFFLOADS,
    MLX5_CAP_ODP,
    MLX5_CAP_ATOMIC,
    MLX5_CAP_ROCE,
    MLX5_CAP_IPOIB_OFFLOADS,
    MLX5_CAP_IPOIB_ENHANCED_OFFLOADS,
    MLX5_CAP_FLOW_TABLE,
    MLX5_CAP_ESWITCH_FLOW_TABLE,
    MLX5_CAP_ESWITCH,
    MLX5_CAP_RESERVED,
    MLX5_CAP_VECTOR_CALC,
    MLX5_CAP_QOS,
    MLX5_CAP_FPGA,
    MLX5_CAP_EMULATION = 0x10, /* NVMx specific */
};

enum {
    MLX5_NO_INLINE_DATA = 0x0,
    MLX5_INLINE_DATA32_SEG = 0x1,
    MLX5_INLINE_DATA64_SEG = 0x2,
    MLX5_COMPRESSED = 0x3,
};

enum devx_event_type {
    DEVX_EVENT_GENERAL_OBJECT = 0x27,
};

struct mlx5_eqe_comp {
	__be32	reserved[6];
	__be32	cqn;
};

union ev_data {
	__be32				   raw[7];
	struct mlx5_eqe_comp   comp;
};

struct mlx5_eqe {
	u8		        rsvd0;
	u8		        type;
	u8		        rsvd1;
	u8		        sub_type;
	__be32		    rsvd2[7];
	union ev_data	data;
	__be16		    rsvd3;
	u8		        signature;
	u8		        owner;
};

enum {
    MLX5_CQ_SET_CI  = 0,
    MLX5_CQ_ARM_DB  = 1,
};

typedef struct devx_cq {
    devx_obj_t  super;
    uint32_t    ci;
    void        *uar_ptr;
} devx_cq_t;

typedef struct devx_uar {
    struct devx_obj_handle *uar;
    uint32_t               uar_id;
    void                   *uar_ptr;
} devx_uar_t;


/* worker represents a polling engine */
typedef struct {
    devx_ctx_t *devx;
    devx_cq_t  tx_cq;
    devx_cq_t  rx_cq;
    devx_uar_t uar;
} devx_worker_t;

struct devx_vtunnel {
    devx_obj_t   super;
    uint16_t     vhca_id;
    uint16_t     vtun_id;
    uint32_t     dma_rkey;
    devx_ctx_t   *sf;
};

typedef struct devx_qp {
    devx_obj_t      super;
    uint8_t         *sq;
    uint32_t        sqi;
    uint8_t         *rq;
    uint32_t        rqi;
    void            *uar_ptr;
    struct devx_qp  *shadow_qp;
    int             pd_id;
    devx_ctx_t      *devx;//device context (PF/SF/ECPF)
    devx_worker_t   *w;// polling engine for regular (non-offloaded) QPs
} devx_qp_t;

typedef struct {
    struct devx_obj_handle *memh;
    struct devx_obj_handle *objh;
    uint32_t                id;
    uint32_t                mkey;
} devx_mr_t;

struct devx_event {
	__u16 obj_type;
	__u32 obj_id;
	__u32 event_type;
};

int devx_get_async_event(void *ctx, struct devx_event *ev);

/*
 * arm cq
 */
void devx_arm_cq(devx_cq_t *cq);

/**
 * poll cq N times
 * 
 * return pointer to the cqe or NULL 
 */
struct mlx5_cqe64 *devx_poll_cq(devx_cq_t *cq, unsigned n);

void devx_rdma_send(devx_qp_t *qp, int rdma_opcode,
        uint32_t lmkey, void *laddr, uint32_t rmkey, void *raddr, size_t size);

/* zero length rdma write with immediate data */
void devx_write_imm(devx_qp_t *qp, uint32_t imm_data);

/* post send op */
void devx_send(devx_qp_t *qp, uint32_t mkey, void *addr, size_t size); 

/* zero length send with immediate data */
void devx_send_imm(devx_qp_t *qp, uint32_t imm_data);

/* post a single recv */
void devx_recv(devx_qp_t *qp, uint32_t mkey, void *addr, size_t size);
/* ring rx doorbell n times */
void devx_ring_rx_db(devx_qp_t *qp, int n);

/*
* managment
* thread safety: requires external lock (todo: internal lock)
*/
int devx_init_dev(const char *dev_name, devx_ctx_t *ctx);
void devx_reset_dev(devx_ctx_t *ctx);

int devx_init_emu_dev(devx_ctx_t *ctx);
void devx_reset_emu_dev(devx_ctx_t *ctx);

/* VHCA tunnel */
int devx_init_tunnel_dev(devx_ctx_t *manager_ctx, devx_ctx_t *tunnel,
                         uint16_t vhca_id);
void devx_reset_tunnel_dev(devx_ctx_t *tunnel);

/* initialize a polling engine:
 * - uar, cqs (rx, tx) and eq
 */
int  devx_worker_init(devx_ctx_t *ctx, devx_worker_t *w);
void devx_worker_reset(devx_worker_t *w);

/* memory registration */
int  devx_reg_mr(devx_ctx_t *ctx, void *buff, size_t size, devx_mr_t *mr);
void devx_dereg_mr(devx_ctx_t *ctx, devx_mr_t *mr);

/* create PD */
int devx_alloc_pd(devx_ctx_t *devx_ctx, devx_obj_t *pd);
/* destroy PD */
void devx_free_pd(devx_ctx_t *devx_ctx, devx_obj_t *pd);

/* query vport mac */
int devx_query_vport_mac(devx_ctx_t *ctx, uint16_t vport, int is_other, uint8_t *vport_mac);
/* query vport lid */
int devx_query_vport_lid(devx_ctx_t *ctx, uint16_t vport, int is_other, uint16_t *vport_lid);

static inline int devx_is_on_ib(devx_ctx_t *ctx)
{
    return ctx->type == MLX5_CAP_PORT_TYPE_IB;
}

/**
 * Check if device has emulation capabilities,
 * like emulation manager for example.
 *
 * Also return tunneled_cmds in case of success (optional).
 */
int devx_emu_caps_support(devx_ctx_t *devx, uint64_t *tunneled_cmds);

/**
 * initialize cq.
 * If eq is not NULL send events to it.
 */
int  devx_cq_init(devx_worker_t *w, devx_cq_t *cq);
void devx_cq_reset(devx_worker_t *w, devx_cq_t *cq);

static inline struct mlx5_cqe64 *devx_worker_poll_tx(devx_worker_t *w, unsigned n) 
{
    return devx_poll_cq(&w->tx_cq, n);
}

static inline struct mlx5_cqe64 *devx_worker_poll_rx(devx_worker_t *w, unsigned n) 
{
    return devx_poll_cq(&w->rx_cq, n);
}

#define DEVX_QP_SQ_PSN 0x4242
#define DEVX_QP_RQ_PSN 0x4242

struct ibv_qp_attr;

int devx_qp_init(devx_ctx_t *devx, devx_worker_t *w, devx_qp_t *qp);
int devx_qp_modify_to_init(devx_qp_t *qp, struct ibv_qp_attr *attr,
                           int ah_mask);
int devx_qp_modify_to_rtr(devx_qp_t *qp, struct ibv_qp_attr *qp_attr,
                          int ah_mask, bool emu_enabled, bool on_behalf_rgid);
int devx_qp_modify_to_rts(devx_qp_t *qp, struct ibv_qp_attr *attr,
                          int ah_mask);

void devx_qp_reset(devx_qp_t *qp);

/* ROCE addr table management */
int devx_roce_addr_get(devx_ctx_t *ctx, int gid_idx, void **roce_addr_p);
int devx_roce_addr_set(devx_ctx_t *ctx, void *roce_addr, int gid_idx);
void devx_roce_addr_free(void *roce_addr);
/* dump to the stdout */
void devx_roce_addr_dump(void *roce_addr);

/* Flow tables. Note that they can only be created over
 * the VHCA tunnel.
 */
typedef struct {
    struct devx_obj_handle *objh;
    uint32_t                table_id;
    uint32_t                table_type;
} devx_ft_t;

typedef struct {
    devx_ft_t              *ft;
    struct devx_obj_handle *objh;
    uint32_t                group_id;
    uint32_t                start_flow_idx;
    uint32_t                end_flow_idx;
    uint8_t                *bitmap;
} devx_fg_t;

typedef struct {
    devx_fg_t              *fg;
    struct devx_obj_handle *objh;
    uint32_t                flow_id;
} devx_fte_t;

enum {
    MLX5_FT_NIC_RX_RDMA = 0x7,
    MLX5_FT_NIC_TX_RDMA = 0x8
};

enum {
    MLX5_FT_MISS_DEFAULT = 0x0,
    MLX5_FT_MISS_FT      = 0x1,
    MLX5_FT_MISS_SD      = 0x2, /* switch domain */
};

#define DEVX_FT_ROOT_LEVEL 5
#define DEVX_FT_LOG_SIZE   10

/* flow actions */
enum {
    MLX5_FTE_ALLOW    = (1<<0),
    MLX5_FTE_DROP     = (1<<1),
    MLX5_FTE_FWD_DEST = (1<<2),
    MLX5_FTE_COUNT    = (1<<3)
};

enum {
    MLX5_FTE_DEST_VPORT   = 0x0,
    MLX5_FTE_DEST_FT      = 0x1,
    MLX5_FTE_DEST_TIR     = 0x2,
    MLX5_FTE_DEST_QP      = 0x3,
    MLX5_FTE_DEST_VHCA_RX = 0x4,
    MLX5_FTE_DEST_VHCA_TX = 0x5,
};

int devx_ft_init(devx_ctx_t *devx_ctx, uint32_t table_type, devx_ft_t *ft);
int devx_ft_reset(devx_ctx_t *devx_ctx, devx_ft_t *ft);
int devx_ft_set_root(devx_ctx_t *devx_ctx, devx_ft_t *ft);

/* direct verb flow rules */
typedef struct {
    struct mlx5dv_flow_matcher *matcher;
    struct ibv_flow            *flow;
} devx_flow_rule_t;

/* redirect rx rdma flow to the given qp to the devx flow table */
int devx_flow_rule_rdma_rx_to_ft(devx_ctx_t *devx_ctx, devx_ft_t *ft,
                                 uint32_t qpn, devx_flow_rule_t *rule);

void devx_flow_rule_reset(devx_flow_rule_t *rule);

/* add flow group to the flow table */
int devx_fg_init(devx_ctx_t *devx_ctx, devx_ft_t *ft, uint32_t start_idx,
                 uint32_t end_idx, uint8_t match_select, void *match_criteria,
                 devx_fg_t *fg);
int devx_fg_reset(devx_ctx_t *devx_ctx, devx_fg_t *fg);

/* add flow table entry to the group, single dest/action only */
int devx_fte_init(devx_ctx_t *devx_ctx, devx_fg_t *fg, uint16_t action,
                  uint8_t dest_type, uint32_t dest_id, void *match_value,
                  devx_fte_t *fte);
int devx_fte_reset(devx_ctx_t *devx_ctx, devx_fte_t *fte);

/* helper functions to build matches and match criteria */
/* dest qp num */
void devx_fte_match_crit_dqpn(void *match_crit);
void devx_fte_match_val_dqpn(void *match_val, uint32_t qpn);
/* source  qp num */
void devx_fte_match_crit_sqpn(void *match_crit);
void devx_fte_match_val_sqpn(void *match_val, uint32_t qpn);
/* match on the ctx port mac */
void devx_fte_match_crit_dmac(void *match_crit);
int devx_fte_match_val_dmac(devx_ctx_t *ctx, void *match_val);

/* memory barriers */

#define devx_compiler_fence() asm volatile(""::: "memory")

#if defined(__x86_64__)

#define devx_memory_bus_fence()        asm volatile ("mfence"::: "memory")
#define devx_memory_bus_store_fence()  asm volatile ("sfence" ::: "memory")
#define devx_memory_bus_load_fence()   asm volatile ("lfence" ::: "memory")

#define devx_memory_cpu_fence()        devx_compiler_fence() 
#define devx_memory_cpu_store_fence()  devx_compiler_fence() 
#define devx_memory_cpu_load_fence()   devx_compiler_fence() 


#elif defined(__aarch64__)

#define devx_memory_bus_fence()        asm volatile ("dsb sy" ::: "memory")
#define devx_memory_bus_store_fence()  asm volatile ("dsb st" ::: "memory")
#define devx_memory_bus_load_fence()   asm volatile ("dsb ld" ::: "memory")

#define devx_memory_cpu_fence()        asm volatile ("dmb ish" ::: "memory")
#define devx_memory_cpu_store_fence()  asm volatile ("dmb ishst" ::: "memory")
#define devx_memory_cpu_load_fence()   asm volatile ("dmb ishld" ::: "memory")

#else
# error "Unsupported architecture"
#endif

#endif
