
/** @file mlx_dev_emu.h
 * Mellanox Device Emulation SDK.
 *
 * ## Overview
 *
 * This library allows emulating a PCI device using Mellanox BlueField-based SmartNIC.
 * It is intended for use by an application running on the SmartNIC's internal processor,
 * and provides the needed functionality for that application to act as a PCI
 * device toward the host where the SmartNIC is located.
 * This application is called the "emulation software" in this API doc.
 *
 * The supplied functionality includes interacting with emulated registers presented to
 * the host, getting notified of doorbells or other actions initiated by the
 * host, DMAing data between host memory and internal processor memory and
 * signaling the host with interrupt.
 *
 * The API is made of several layers. A basic layer that allows emulating
 * generic devices (but not all possible devices!), and layers that are more
 * efficient for specific types of devices, such as NVMe emulation. These APIs
 * use specific offloads provided by the SmartNIC to allow such efficiency.
 * 
 * ## Devices Model
 *
 * SmartNIC is preconfigured (using Mellanox nvconfig tool) to present emulated
 * device PFs and VFs. For each emulated PF presented to the host, an RDMA PF
 * is presented to the SmartNIC internal processor. The emulated PF and all its
 * VFs are managed and serviced using that RDMA PF inside the SmartNIC.
 *
 * ## Working with emulated devices
 *
 * An emulated device has a BAR with registers space. The emulation software
 * using the SDK can communicate with the host using those registers.
 * The registers act as a memory: both host and emulation software reads/writes
 * to them act as reads/writes to shared memory space.
 *
 * It is expected that the device defines an efficient way for communications
 * with the host using queues. Each such queue is made of a piece of host
 * memory that acts like a queue and a doorbell. Host should be writing
 * elements to the queue, and then write the last producer index to the
 * doorbell. The SDK will facilitate transferring those queue posts to the
 * emulation software.

 * ## Threads Model
 *
 * A worker context is used to describe a single thread. This thread should be
 * responsible for a group of emulated queues created for a specific emulated
 * device.
 * The thread should periodically call the progression function in this SDK
 * so that the different callbacks can be called based on available events from
 * the queues bound to the thread.
 * For events driven model, a file descriptor can be used to poll/epoll/select
 * on, which will wake up the thread when a new event is waiting. The thread
 * is expected to run the progression function when waking up.
 *
 * ## Types of emulation
 *
 * The most basic emulation type delivers the doorbells that the host is
 * writing to the emulation software. The emulation software is then expected
 * to use the DMA capabilities in the SDK to bring the queue elements, parse
 * them, act on them, potentially DMA more data from/to host, and eventually
 * DMA a completion item to the emulated completion queue and signal an
 * interrupt. In this mode, the completion queue is another emulated queue
 * that repors doorbells when the host consumes a completion item.
 *
 * More advanced modes of emulation adds knowledge on the queues and the
 * elements posted to them, and make it more efficient for the emulation
 * software.
 *
 * ## Flow
 *
 * The general flow of calls to emulate a device is:
 * 1. Call mlx_emu_get_version() to get library version and capabilities.
 * 2. Call mlx_emu_dev_query() to get device emulation capabilities and status of a specific RDMA NIC function.
 * 3. Call mlx_emu_dev_open() to "open" a specific emulated device function and begin working with it.
 * 4. Call mlx_emu_worker_create() to create a worker thread context.
 * 5. Optionally call mlx_emu_worker_fd() to get a file descriptor to sleep on (not yet supported). Otherwise work in polling mode.
 * 6. Call mlx_emu_worker_progress() periodically or when file descriptor is ready for read.
 * 7. Call mlx_emu_bar_read() and mlx_emu_bar_write() to communicate with the host using device registers.
 * 8. Call mlx_emu_q_create() to create an emulated queue. SDK will now track this queue (doorbell rings or more).
 * 9. Handle the callback whenever a new work is posted by the host to the emulated queue.
 * 10. Use mlx_emu_q_dma_read_bcopy() and mlx_emu_q_dma_write_bcopy() to DMA queue elements and other data between
 *     emulation software and host memory.
 * 11. Call mlx_emu_q_interrupt() to signal an interrupt to the host. i.e, after DMA of a completion element.
 *
 */

#ifndef MLX_DEV_EMU_H
#define MLX_DEV_EMU_H

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>

#include "devx_verbs.h"

/**
 * Status return codes.
 */
enum {
    MLX_EMU_OK  =  0,
    MLX_EMU_ERR = -1,
    MLX_EMU_NOTSUPPORTED = -2,
};


#define MLX_EMU_BAR_CB_TRIGGER (100 * 1000 / CLOCKS_PER_SEC)	/* 100 msec */

#define MLX_DEV_EMU_LOCK_FD_PREFIX "/var/lock"

#define MLX_DEV_EMU_VER_MAJOR   0
#define MLX_DEV_EMU_VER_MINOR   0
#define MLX_DEV_EMU_VER_SUB     1

/**
 * Callback function indicating registers space of emulated device was written by host.
 *
 * Should be supplied when calling mlx_emu_dev_open().
 * Called when host has written to the bar registers space, which is on BAR 0
 * from the beginning and up to mlx_emu_dev_attr_t.bar_reg_size supplied to mlx_emu_dev_open().
 * Since registers are per device, this callback is always called from the
 * context for the first worker.
 */
typedef void (*mlx_emu_bar_write_cb_t)(void *ctx);

/**
 * Callback function indicating that FLR (function level reset) was
 * triggered.
 *
 * @param[in] status will be MLX_EMU_FLR_START when FLR is detected and
 *            MLX_EMU_FLR_COMPLETED when library is ready to be used
 */
enum {
    MLX_EMU_FLR_START,
    MLX_EMU_FLR_COMPLETED,
};

typedef void (*mlx_emu_flr_cb_t)(void *ctx, int status);

/**
 * Device Emulation Attributes.
 *
 * Describes basic attributes and capabilities of the device emulation.
 */
typedef struct {
    struct {
        uint16_t major;
        uint16_t minor;
        uint16_t sub;
    } v;		/**< version number */
    uint64_t caps;      /**< capabilitis mask, not yet implemented */
    size_t   max_bcopy; /**< max size that can be transfered via bcopy read/write */
} mlx_emu_ver_t;

/**
 * Get library version and capabilities.
 * @param[out] attr Pointer to API attributes
 */
int mlx_emu_get_version(mlx_emu_ver_t *attr);

typedef struct mlx_emu_dev mlx_emu_dev_t; /**< Emulated device opaque context */

struct mlx_emu_dev {
    devx_ctx_t             ecpf;
    uint8_t                num_emu_pfs;

    uint64_t               tunneled_cmds; /* mask of MLX5_VHCA_TUNNEL_CMD_* */
    size_t                 bar_size;

    uint16_t               db_base;
    mlx_emu_bar_write_cb_t bar_write_cb;
    uint16_t               bar_reg_size;
    mlx_emu_flr_cb_t       flr_cb;
    void                   *cb_ctx;
    uint32_t               dma_rkey;
    int                    lock_fd;
    char                   lock_fd_path[32];
};

/**
 * Emulated device attributes.
 *
 * Supplied when opening a speicifc emulated device function, and includes
 * parameters. 
 */
typedef struct {
    uint16_t function;       /**< function number of the emulated device to open */
    uint16_t bar_reg_size;   /**< actual size of the registers space in the BAR */
    uint16_t db_base;        /**< offset from the beginnig of BAR 0 where doorbells start */
    uint64_t caps;	     /**< required capabilities */
    /* lock policy ? */
    mlx_emu_bar_write_cb_t    bar_write_cb; /**< a callback function to be called whenever the registers space has been written by the host */
    void                      *cb_ctx;
    mlx_emu_flr_cb_t          flr_cb;
} mlx_emu_dev_attr_t;

/**
 * Device Emulation capabilities and attributes of specific RDMA device.
 *
 */
typedef struct {
    uint16_t max_functions;     /**< number of emulated functions that are presented to the host */
    uint32_t max_namespaces;    /**< max number of namespaces support per SQ */
    uint16_t max_bar_reg_size;  /**< max size of the bar register area */
    uint32_t max_emulated_cq;   /**< max number of CQs that can be emulated */
    uint32_t max_emulated_sq;   /**< max number of SQs that can be emulated */
    uint64_t caps;              /**< emulation capabilities  */
} mlx_emu_dev_query_attr_t;

/**
 * Query device emulation capabilities and status of a specific RDMA device.
 *
 * Emulation software should run this query per every RDMA device available
 * to the SmartNIC cpu prior to initializing the device. Those RDMA devices
 * that return successful value of MLX_EMU_OK are in charge of emulated
 * devices, and the emulation software should work with them to service the
 * emulation. The returned dev_caps indicates how many such emulated functions
 * this RDMA device is in charge of and the emulation software should service.
 *
 * Other RDMA devices either do not present any emulated devices or don't
 * support emulation at all.
 *
 * @param[in] dev_name The name of the RDMA device to query, such as "mlx5_2".
 * @param[out] dev_caps The emulation capabilities and status avaiable by the RDMA device.
 * @retval MLX_EMU_OK Device emulation is supported by this RDMA device
 * @retval MLX_EMU_NOTSUPPORTED Device emulation is not supported by this RDMA device
 * @retval MLX_ERR Error while querying the RDMA device
 */
int mlx_emu_dev_query(const char *dev_name, mlx_emu_dev_query_attr_t *dev_caps);
int mlx_emu_dev_caps(devx_ctx_t *devx, mlx_emu_dev_query_attr_t *dev_caps);

/**
 * Open a specific device emulation manager to begin working with it.
 * Only a privileged mlx function has this capability.
 *
 * Emulation software should call this for each of the emulated functions
 * in order to begin working with them.
 *
 * This call will check the supplied attributes and then allocate a new context 
 * indentifying the emulated device function for the rest of this API.
 *
 * @param[in] dev_name RDMA device name, such as "mlx5_2".
 * @param[in] attr Parameters for the emulated device.
 * @param[out] if_name Network device name that is associated with dev_name, if successful.
 *
 * @retval device context allocated and returned, if successful. Otherwise NULL.
 *
 */
mlx_emu_dev_t *mlx_emu_dev_open(mlx_emu_dev_attr_t *attr);

/**
 * Close an device emulation manager context.
 * 
 * Close and deallocate emulated device context, previously allocated using mlx_emu_dev_open()
 *
 */
int mlx_emu_dev_close(mlx_emu_dev_t *emu_manager);

#endif
