/*
 * Mellanox Device Emulation Object.
 * This context is used to read/write the registers BAR of an emulated device.
 * In case of writing to these registers, a Device Emulation Event will be generated.
 * The size of the Bar registers is reported in HCA_CAP.Device_Emulation_CAPs.registers_size.
 * Note that read and write of this context is allowed only for device emulation manager.
 */
#ifndef _MLX_DEVICE_EMULATION_OBJ_H
#define _MLX_DEVICE_EMULATION_OBJ_H

#include <stdint.h>

#include "devx_verbs.h"

typedef struct mlx_emu_dev_obj {
    struct devx_obj_handle *emu;
    devx_ctx_t             *ecpf;
    void                   *bar;
    size_t                  bar_size;
    int                     bar_fetched;
    uint16_t                emu_vf_id;
    uint8_t                 emu_dev_id; /* emulation device id as supplied by FW */
    uint16_t                emu_vid;
} mlx_emu_dev_obj_t;

/**
 * Init device emulation object
 */
int mlx_emu_dev_obj_init(devx_ctx_t *devx_ctx, mlx_emu_dev_obj_t *dev_emu_obj,
                         uint16_t emu_vf_id, size_t bar_size);
/**
 * Reset device emulation object
 */
void mlx_emu_dev_obj_reset(mlx_emu_dev_obj_t *dev_emu_obj);

/**
 * Read host BAR.
 *
 * When the host writes to BAR registers, they are kept by the device until the
 * emulation software uses this call to fetch the latest values.
 *
 * It is possible to read a single register, a bunch of contiguous registers
 * or even the entire registers space at once.
 *
 * @param[in] dev_ctx the emulated device context.
 * @param[in] buf a local buffer to copy the registers values to.
 * @param[in] bar_addr the address to read from the start of the BAR.
 * @param[in] len the length to read.
 */ 
int mlx_emu_bar_read(mlx_emu_dev_obj_t *dev_emu_obj, void *buf,
                     uint32_t bar_addr, unsigned len);

/**
 * Write to host BAR.
 *
 * When the host reads BAR registers, it gets a value that was previously written
 * to that register either by the host itself or by the emulation software
 * using this call. 
 *
 * It is possible to write a single register, a bunch of contiguous registers
 * or even the entire registers space at once.
 *
 * @param[in] dev_ctx the emulated device context.
 * @param[in] buf a local buffer to copy the registers values from.
 * @param[in] bar_addr the address to write to from the start of the BAR.
 * @param[in] len the length to write.
 */ 
int mlx_emu_bar_write(mlx_emu_dev_obj_t *dev_emu_obj, void *buf,
                      uint32_t bar_addr, unsigned len);

#endif
