#ifndef MLX_EMU_QUEUE_H
#define MLX_EMU_QUEUE_H

#include <stdint.h>

#include "devx_verbs.h"
#include "nvme_mlx5_ifc.h"

typedef struct mlx_emu_q mlx_emu_q_t; // General emulated queue context
typedef struct mlx_emu_q_attr mlx_emu_q_attr_t;
typedef struct mlx_emu_q_bbuf mlx_emu_q_bbuf_t; // General emulated queue context
typedef struct mlx_nvme_cq mlx_nvme_cq_t;
typedef struct mlx_nvme_sq mlx_nvme_sq_t;

struct mlx_emu_cb_base;

struct mlx_emu_q_bbuf {
    uint8_t                *bbuf;
    struct ibv_mr          *mr;
};

typedef struct {
    uint32_t    device_emulation_id;
    uint8_t     offload_type;
    uint8_t     network_state;
    uint32_t    qpn;
    uint16_t    qpn_vhca_id;
    uint8_t     log_nvme_page_size;
    uint16_t    msix_vector;
    uint64_t    nvme_base_addr;
    uint8_t     nvme_log_entry_size;
    uint16_t    nvme_num_of_entries;
    uint8_t     max_transcation_size;
    uint32_t    nvme_doorbell_offset;
    uint8_t     cq_period; // for CQ only
    uint8_t     cq_max_count; // for CQ only
    uint16_t    nvme_cq_id; // for SQ only
    uint32_t    pd_id;  // for SQ only
    uint64_t    fields_to_modify;
} mlx_nvme_queue_attr_t;

/**
 * Attributes of an emulated queue.
 *
 * An emulated queue refers to a memory region in host memory acting as a queue
 * with a doorbell address where the host reports producer index.
 * When creating an emulated queue different information about the queue is
 * needed  depending on the mode of emulation requested. A different
 * callback is provided to deliver the notifications of submittions to the
 * queue.
 */
struct mlx_emu_q_attr {
     uint16_t                db_base;
    uint32_t                 db_offset;  /**< doorbell offset relative to the device 
                                           doorbell base. Always needed */
    uint16_t                 max_length; /**< Max queue length. Always needed. */
    uint32_t                 flags;       /**< See @ref enum mlx_emu_q_flags. Always needed */
    int                      msi_x;	     /**< msix interrupt to generate when requested.
                                           Should be set when this queue is a 
                                           a completion queue */
    uint64_t				 base_addr;  /**< sqe base physical address */
    uint8_t					 log_entry_size; /**< log(base 2) of the sqe entry size. */
    uint8_t					 log_page_size; /**< Log(base 2) of the page size host is
     	 	 	 	 	 	 	 	 	     working with the NVMe device in granularity of 4KB.
     	 	 	 	 	 	 	 	 	      i.e. page size in bytes is 2^(12+log_nvme_page_size). */
    uint8_t 				 max_transcation_size; /**< Device will verify that each SQE does not exceeds
    												this size in granularity of 2^log_nvme_page_size. */
    uint32_t                 ext_qp_num; /**< if non zero and MLX_EMU_Q_EXTERNAL_QP flag is set
                                           SDK will use the qp to send all 
                                           emulation related data such as doorbells, NVMoF
                                           command capsules etc. The caller is responsible for
                                           the qp lifecyle management. E.x. connecting to the
                                           remote qp, handling errors, receiving data etc...
                                           The callbacks normally used to deliver commands 
                                           (such as mlx_emu_q_nvmf_cmd_cb_t) will never be called.
                                           When external QP is used, there is no sense in having more
                                           than one worker that will handle async events and registers 
                                           events, so all queues can be created using the same worker context. 
                                          */
    mlx_emu_q_t             *cq;         /**< If the queue is a submission queue it should point to
                                              a valid completion queue */ 
    struct mlx_emu_cb_base  *cb;         /**< mode of emulation and related callbacks
                                           and info */
    void                    *ucontext;   /**< user context */
    uint8_t                 cq_period;
    uint8_t                 cq_max_count;
};

struct mlx_emu_q {
    mlx_emu_q_attr_t       attrs;
    mlx_emu_q_bbuf_t       *bbuf;
    uint32_t               dma_rkey;
    devx_ctx_t             *devx_ctx;

    uint16_t               emu_vf_id;
    uint16_t               sf_vhca_id;

    bool active;
};

/* Mellanox NVMe CQ object */
struct mlx_nvme_cq {
    mlx_emu_q_t             emu_q;
    struct devx_obj_handle  *objh;
    uint32_t                id;
};

/* Mellanox NVMe SQ object */
struct mlx_nvme_sq {
    mlx_emu_q_t             emu_q;
    struct devx_obj_handle  *objh;
    uint32_t                id;
    uint32_t                qpn;
    /* PF side rules */
    devx_fg_t               *fg_tx;
    devx_fte_t              fte_tx;
    devx_fg_t               *fg_rx;
    devx_fte_t              fte_rx;
    /* ECPF/SF rules... */
    devx_flow_rule_t        rdma_rx_rule;
    devx_ft_t               *ft_sf_rx;
};

static inline mlx_nvme_cq_t *to_mlx_nvme_cq(mlx_emu_q_t *emu_q)
{
    return container_of(emu_q, mlx_nvme_cq_t, emu_q);
}

static inline mlx_nvme_sq_t *to_mlx_nvme_sq(mlx_emu_q_t *emu_q)
{
    return container_of(emu_q, mlx_nvme_sq_t, emu_q);
}

/**
 * A doorbell callback.
 * Called when host has written to the doorbell address of the emulated
 * queue.
 * Called from the context of the worker that the emulated queue belongs to.
 *
 * @param[in] q the emulated queue that is the source of this doorbell.
 * @param[in] db_addr the doorbell address that was written to.
 * @param[in] db_val the doorbell value (producer index) that was written.
 */
typedef void (*mlx_emu_q_db_cb_t)(mlx_emu_q_t *q, uint32_t db_addr, uint32_t db_val);


/**
 * NVME command to NVMF command capsule conversion callback
 */
typedef void (*mlx_emu_q_nvmf_cmd_cb_t)(mlx_emu_q_t *q, uint32_t db_addr, void *cmd, uint32_t len);

/**
 * NVME command callback
 */
typedef void (*mlx_emu_q_nvme_cmd_cb_t)(mlx_emu_q_t *q, uint32_t db_addr, void *cmd, uint32_t len);

/**
 * Queue emulation types.
 */
enum {
    MLX_EMU_CB_TYPE_DB,	                /**< Only pass doorbell notifications */
    MLX_EMU_CB_TYPE_NVME,               /**< Pass raw NVME SQE capsule */
    MLX_EMU_CB_TYPE_NVME2NVMF,          /**< Convert NVME command to NVMF command capsule */
};

/**
 * A basic structure for queue emulation callbacks.
 *
 * Should not be used directly. Holds nothing but the type of emulation.
 * Use the mlx_emu_xxx_mode_db_t structures instead.
 */
typedef struct mlx_emu_cb_base {
   uint8_t emu_cb_type;	/**< Emulation type, from the MLX_EMU_CB_TPYE_xxx enumbs */
} mlx_emu_cb_base_t; 

/**
 * A queue emulation of type doorbell callbacks structure (DBR to DBR mode)
 */
typedef struct {
    mlx_emu_cb_base_t   super;	/**< the basic structure which holds the emulation type */
    mlx_emu_q_db_cb_t   doorbell_write_cb;	/**< a callback for delivering a doorbell notification */
} mlx_emu_doorbell_mode_cb_t;

/**
 * NVME to NVMF queue emulation (DBR [to SQE] to CC mode)
 */
typedef struct {
    mlx_emu_cb_base_t        super;         
    mlx_emu_q_nvmf_cmd_cb_t  nvmf_command_cb;  /**< callback for delivering a NVME command converted 
                                                    to a NVMF command capsule */
} mlx_emu_nvmf_mode_cb_t;

/**
 * NVME queue emulation (DBR to SQE mode)
 */
typedef struct {
    mlx_emu_cb_base_t        super;         
    mlx_emu_q_nvme_cmd_cb_t  nvme_command_cb;  /**< callback for delivering a NVME command */ 
} mlx_emu_nvme_mode_cb_t;

/**
 * Device Emulation queue types
 */
enum mlx_emu_q_flags {
    MLX_EMU_Q_NONE        = 0,      /**< A default queue */
    MLX_EMU_Q_CQ          = (1<<0), /**< The queue represents a completion queue */
    MLX_EMU_Q_SQ          = (1<<1), /**< The queue is a submission queue */
    MLX_EMU_Q_EXTERNAL_QP = (1<<2)  /**< The queue uses an external QP to transfer data */
};

enum {
    MLX_EMU_MODIFY_SQ_QPN   = (1 << 0),
    MLX_EMU_MODIFY_SQ_STATE = (1 << 1),
};

/**
 * Create an emulated queue.
 *
 * This call sets up an emulated queue to track. The SDK will notify the
 * emulation software of changes happening in the emulated queue, via the
 * callbacks supplied.
 *
 * Depending on the type of emulation requested in the supplied attributes,
 * different information is needed about the emulated queue, and a different
 * notification callback will be called.
 *
 * Supported queue emulation modes are:
 * 1. Doorbell. In this mode, SDK will notify the emulation software whenever
 *    a doorbell was written by the host to the specified doorbell address. The
 *    notification includes the new doorbell value. The emulation software
 *    callback is then required to fetch the actual queue emelents that were
 *    posted by the host based on this value. In this mode the same call and 
 *    mode should be used for the emulated completion queue.
 *
 * As soon as this call returns the emulated queue tracking begins, and calls
 * to mlx_emu_worker_progress() will dispatch the supplied callback.
 *
 * @param[in] worker The worker context this queue handling will belong to.
 *       Events generated by this queue will issue callbacks only
 *       when this worker context is called with mlx_emu_worker_progress().
 * @param[in] attr Attributes for the emulated queue.
 * @param[out] q_p A queue context representing this queue for further
 *       reference in this API.
 */
int mlx_emu_q_create(devx_ctx_t *devx_ctx, mlx_emu_q_bbuf_t *bbuf,
                     mlx_emu_q_attr_t *attr, mlx_emu_q_t *q, uint32_t dma_rkey,
                     uint16_t emu_vf_id);

int mlx_nvme_sq_modify(mlx_nvme_sq_t *sq,
                       mlx_nvme_queue_attr_t *q_attr_in);

/**
 * Attributes that can be modified by the @ref mlx_emu_q_modify()
 */
enum mlx_emu_q_modify_mask {
    MLX_EMU_Q_MODIFY_EXTERNAL_QP = 1 /** < Change an external QP number */
};

/**
 * Modify an emulated queue (Not implemented)
 *
 * Change queue attributes. When changing external qp number it is up to
 * the caller to gurantee that data loss does not happen. 
 * @param[in] attr_mask a bitwise mask of queue atrributes @ref enum mlx_emu_q_modify_mask
 *            that should be modified. At the moment only 
 *            MLX_EMU_Q_MODIFY_EXTERNAL_QP is supported.
 *
 * @retval MLX_EMU_NOTSUPPORTED if queue can not be modified at all
 * @retval MLX_EMU_ERR if there was some kind of error
 */
int mlx_emu_q_modify(mlx_emu_q_t *q, mlx_emu_q_attr_t *attr, uint32_t attr_mask);

/**
 * Query current queue attributes
 */
int mlx_emu_q_query(mlx_emu_q_t *q, mlx_emu_q_attr_t *attrs);

/**
 * Get a user supplied queue context
 *
 * @retval a context supplied during the queue creation
 */
void *mlx_emu_q_context(mlx_emu_q_t *q);

/**
 * Destroy an emulated queue.
 *
 * Destroy an emulated queue previously created using mlx_emu_q_create().
 */
int mlx_emu_q_destroy(mlx_emu_q_t *q);

/**
 * Send interrupt to the host.
 *
 * Use the given queue to send an interrupt to the host. The MSIX interrupt
 * vector that was configured with the queue upon its creation will be used.
 *
 * @param[in] q The queue that should generate an interrupt
 */
void mlx_emu_q_interrupt(mlx_emu_q_t *q);

/**
 * DMA read from host memory.
 * 
 * Read from host memory address and length and copy it to the supplied local buffer.
 * The read is synchronous, which means the call will block untill read is complete.
 * An internal copy of data happens during the call.
 *
 * @param[in] q the emulated queue to use for the DMA.
 * @param[in] dst_buf a destination local buffer to copy the data into.
 * @param[in] srcaddr the physical address of host memory to read from.
 * @param[in] rmkey the remote key for the RDMA (0 means we'll use global key).
 * @param[in] len the length of data to DMA.
 */
void mlx_emu_q_dma_read_bcopy(mlx_emu_q_t *q, void *dst_buf, uint64_t srcaddr,
                              uint32_t rmkey, size_t len);

/**
 * DMA write to host memory.
 * 
 * Write to host memory address and length the supplied local buffer.
 * The write is synchronous, which means the call will block untill write is complete.
 * An internal copy of data happens during the call.
 *
 * @param[in] q the emulated queue to use for the DMA.
 * @param[in] src_buf a source local buffer to copy the data from.
 * @param[in] dstaddr the physical address of host memory to write to.
 * @param[in] rmkey the remote key for the RDMA (0 means we'll use global key).
 * @param[in] len the length of data to DMA.
 */
void mlx_emu_q_dma_write_bcopy(mlx_emu_q_t *q, void *src_buf, uint64_t dstaddr,
                               uint32_t rmkey, size_t len);

/**
 * DMA send to host memory.
 *
 * Send local buffer to host memory
 *
 * @param[in] q the emulated queue to use for the DMA.
 * @param[in] src_buf a source local buffer to copy the data from.
 * @param[in] len the length of data to DMA.
 */
void mlx_emu_q_send_bcopy(mlx_emu_q_t *q, void *src_buf, size_t len);

#endif
