/*
 * Mellanox NVMe emulator controller
 */
#ifndef _MLX_NVME_EMULATOR_H
#define _MLX_NVME_EMULATOR_H

#include <stdint.h>
#include <linux/if_ether.h>

#include "nvme_emu.h"
#include "mlx_dev_emu.h"
#include "mlx_sf.h"
#include "mlx_device_emulation_obj.h"
#include "mlx_emu_queue.h"
#include "mlx_nvme_emu_namespace.h"
#include "devx_verbs.h"

#define MLX_EMU_MAX_BCOPY   8192
#define MLX_NVME_EMU_HCA_RETRY_CNT 100
#define MLX_NVME_EMU_HCA_USEC_WAIT 50000
#define MLX_NVME_MAX_COMPLETIONS 16

typedef struct mlx_nvme_emulator mlx_nvme_emulator_t;
typedef struct mlx_nvme_emu_sq mlx_nvme_emu_sq_t;
typedef struct mlx_nvme_emu_cq mlx_nvme_emu_cq_t;


typedef struct mlx_nvme_emu_qp_loop {
    mlx_nvme_emu_sq_t *sq;
    struct ibv_cq *tx_cq;
    struct ibv_cq *rx_cq;
    struct ibv_qp *qp;
    struct ibv_recv_wr rx_wr[DEVX_RQ_SIZE];
    struct ibv_sge rx_sge[DEVX_RQ_SIZE];

    devx_qp_t     qp_hw;
} mlx_nvme_emu_qp_loop_t;

struct mlx_nvme_emu_sq {
    mlx_nvme_sq_t          sq;
    mlx_emu_q_bbuf_t       rq_bbuf;
    mlx_nvme_emu_qp_loop_t *fw_qp;
    devx_ctx_t             *devx_ctx;
    devx_ctx_t             *sf;
    int                    qid;
    /* Offload attributes */
    devx_qp_t              *offload_qp;
    struct devx_obj_handle *rule;
};

struct mlx_nvme_emu_cq {
    mlx_nvme_cq_t      cq;
    int                qid;
};

/**
 * struct mlx_nvme_emulator - Mellanox NVMe emulation context
 *
 * @emulator:     General NVMe emulation context
 * @lock:         Emulator lock
 * @emu_enabled:  Enabled bit (TODO: move to state machine)
 * @started:      Started bit (TODO: move to state machine)
 * @w:            Emulation worker (TODO: remove this in the future)
 * @emu_manager:  Mellanox Emulation manager context
 * @sqe_mode_cb:  Callbacks for SQE mode emulation
 * @cc_mode_cb:   Callbacks for CC mode emulation
 * @sf:           Mellanox Sub-Function context (will be used for networking and RDMA)
 * @dev_emu_obj:  Mellanox device emulation object
 * @tunnel:       Devx context for VHCA tunnel
 * @ft_rx:        Flow table context for RX root steering rules
 * @fg_rx:        Flow group context for RX root flow tables
 * @fg_rx_miss:   Flow group context for RX miss flow tables
 * @fte_rx_miss:  Flow table entry context for RX miss steering rules
 * @ft_tx:        Flow table context for TX root steering rules
 * @fg_tx:        Flow group context for TX root flow tables
 * @emu_dev_id:   Emulation device id as supplied by FW
 * @emu_vf_id:    Emulation virtual function id as supplied by FW
 */
struct mlx_nvme_emulator {
    nvme_emulator_t              emulator;
    pthread_mutex_t              lock;
    bool                         enabled;
    bool                         started;
    mlx_emu_dev_t                *emu_manager;
    mlx_emu_nvme_mode_cb_t       sqe_mode_cb;
    mlx_emu_nvmf_mode_cb_t       cc_mode_cb;
    mlx_sf_t                     *sf;
    mlx_emu_dev_obj_t            dev_emu_obj;
    devx_ctx_t                   tunnel;
    devx_ft_t                    ft_rx;
    devx_fg_t                    fg_rx;
    devx_fg_t                    fg_rx_miss;
    devx_fte_t                   fte_rx_miss;
    devx_ft_t                    ft_tx;
    devx_fg_t                    fg_tx;
    uint8_t                      emu_dev_id;
    uint16_t                     emu_vf_id;

    uint16_t                     num_queues;
    mlx_nvme_emu_sq_t            *sqs;
    mlx_nvme_emu_cq_t            *cqs;
    clock_t                      last_bar_cb;
    mlx_emu_q_bbuf_t             bbuf;

};

static inline mlx_nvme_emulator_t *to_mlx_nvme_emulator(nvme_emulator_t *emulator)
{
    return container_of(emulator, mlx_nvme_emulator_t, emulator);
}

static inline mlx_nvme_emu_cq_t *to_mlx_nvme_emu_cq(mlx_nvme_cq_t *cq)
{
    return container_of(cq, mlx_nvme_emu_cq_t, cq);
}

static inline mlx_nvme_emu_sq_t *to_mlx_nvme_emu_sq(mlx_nvme_sq_t *sq)
{
    return container_of(sq, mlx_nvme_emu_sq_t, sq);
}

enum {
    MLX_NVME_EMU_SQ_STATE_INIT = 0x0,
    MLX_NVME_EMU_SQ_STATE_RDY,
    MLX_NVME_EMU_SQ_STATE_ERROR,
};

mlx_nvme_emu_qp_loop_t *mlx_nvme_open_qp_loop(mlx_nvme_emulator_t *mdev, mlx_sf_t *sf,
                                              mlx_nvme_emu_sq_t *sq);
void mlx_nvme_close_qp_loop(mlx_nvme_emu_qp_loop_t *qp_loop);
int mlx_nvme_emu_sq_modify(mlx_nvme_emu_sq_t *sq, uint32_t qpn, uint8_t network_state);
int mlx_nvme_emu_init_offload_qp(mlx_nvme_emulator_t *mdev,
                                 devx_qp_t *offload_qp, mlx_nvme_emu_sq_t *sq);
void mlx_nvme_emu_reset_offload_qp(devx_qp_t *offload_qp, mlx_nvme_emu_sq_t *sq);
int mlx_nvme_emu_modify_offload_qp(mlx_nvme_emulator_t *mdev,
                                   devx_qp_t *qp, struct ibv_qp_attr *qp_attr,
                                   int attr_mask);
int mlx_nvme_emu_send_offload_cmd_sync(mlx_nvme_emulator_t *mdev,
                                       devx_qp_t *offload_qp,
                                       mlx_nvme_emu_sq_t *sq, void *data,
                                       int data_len, int retry_count,
                                       int usec_retry, uint8_t *cmd,
                                       void *output, int output_len);

int mlx_nvme_init_hca(mlx_nvme_emulator_t *mdev);
int mlx_nvme_teardown_hca(mlx_nvme_emulator_t *mdev);
int mlx_nvme_enable_hca(mlx_nvme_emulator_t *mdev);
int mlx_nvme_disable_hca(mlx_nvme_emulator_t *mdev);
int mlx_nvme_alloc_tunneled_pd(mlx_nvme_emulator_t *mdev);
void mlx_nvme_free_tunneled_pd(mlx_nvme_emulator_t *mdev);

int mlx_nvme_setup_roce_table(mlx_nvme_emulator_t *mdev);

#endif
