#ifndef _NVME_H
#define _NVME_H

#include "nvme_regs.h"
#include "compiler.h"

#define NVME_VERSION(mjr, mnr, ter) \
	(((uint32_t)(mjr) << 16) | \
	((uint32_t)(mnr) << 8) | \
	(uint32_t)(ter))

#define NVME_NAMESPACE_GLOBAL 0xFFFFFFFF

enum nvme_quirks {
    NVME_QUIRK_ASYNC_EV_CONFIG        = 1 << 0,
    NVME_QUIRK_NS_CHANGE_FORCE_UNMASK = 1 << 1,
};

enum NvmeCstsShift {
    CSTS_RDY_SHIFT      = 0,
    CSTS_CFS_SHIFT      = 1,
    CSTS_SHST_SHIFT     = 2,
    CSTS_NSSRO_SHIFT    = 4,
    CSTS_PP_SHIFT       = 5,
};

enum NvmeCstsMask {
    CSTS_RDY_MASK   = 0x1,
    CSTS_CFS_MASK   = 0x1,
    CSTS_SHST_MASK  = 0x3,
    CSTS_NSSRO_MASK = 0x1,
    CSTS_PP_MASK    = 0x1,
};

enum NvmeCsts {
    NVME_CSTS_READY         = 1 << CSTS_RDY_SHIFT,
    NVME_CSTS_FAILED        = 1 << CSTS_CFS_SHIFT,
    NVME_CSTS_SHST_NORMAL   = 0 << CSTS_SHST_SHIFT,
    NVME_CSTS_SHST_PROGRESS = 1 << CSTS_SHST_SHIFT,
    NVME_CSTS_SHST_COMPLETE = 2 << CSTS_SHST_SHIFT,
    NVME_CSTS_NSSRO         = 1 << CSTS_NSSRO_SHIFT,
};

enum nvme_psdt_value {
    NVME_PSDT_PRP              = 0x0,
    NVME_PSDT_SGL_MPTR_CONTIG  = 0x1,
    NVME_PSDT_SGL_MPTR_SGL     = 0x2,
    NVME_PSDT_RESERVED         = 0x3
};

typedef struct NVME_PACKED nvme_lbaf {
    uint16_t    ms;
    uint8_t     ds;
    uint8_t     rp;
} nvme_lbaf_t;

typedef struct NVME_PACKED nvme_id_ns_t {
    uint64_t    nsze;
    uint64_t    ncap;
    uint64_t    nuse;
    uint8_t     nsfeat;
    uint8_t     nlbaf;
    uint8_t     flbas;
    uint8_t     mc;
    uint8_t     dpc;
    uint8_t     dps;
    uint8_t     res30[74];
    uint8_t     nguid[16];
    uint8_t     eui64[8];
    nvme_lbaf_t lbaf[16];
    uint8_t     res192[192];
    uint8_t     vs[3712];
} nvme_id_ns_t;

#define NVME_ID_NS_NSFEAT_THIN(nsfeat)      ((nsfeat & 0x1))
#define NVME_ID_NS_FLBAS_EXTENDED(flbas)    ((flbas >> 4) & 0x1)
#define NVME_ID_NS_FLBAS_INDEX(flbas)       ((flbas & 0xf))
#define NVME_ID_NS_MC_SEPARATE(mc)          ((mc >> 1) & 0x1)
#define NVME_ID_NS_MC_EXTENDED(mc)          ((mc & 0x1))
#define NVME_ID_NS_DPC_LAST_EIGHT(dpc)      ((dpc >> 4) & 0x1)
#define NVME_ID_NS_DPC_FIRST_EIGHT(dpc)     ((dpc >> 3) & 0x1)
#define NVME_ID_NS_DPC_TYPE_3(dpc)          ((dpc >> 2) & 0x1)
#define NVME_ID_NS_DPC_TYPE_2(dpc)          ((dpc >> 1) & 0x1)
#define NVME_ID_NS_DPC_TYPE_1(dpc)          ((dpc & 0x1))
#define NVME_ID_NS_DPC_TYPE_MASK            0x7
#define NVME_ID_NS_DPS_TYPE_MASK            0x7

enum nvme_sgl_descriptor_type {
    NVME_SGL_TYPE_DATA_BLOCK           = 0x0,
    NVME_SGL_TYPE_BIT_BUCKET           = 0x1,
    NVME_SGL_TYPE_SEGMENT              = 0x2,
    NVME_SGL_TYPE_LAST_SEGMENT         = 0x3,
    NVME_SGL_TYPE_KEYED_DATA_BLOCK     = 0x4,
    NVME_SGL_TYPE_TRANSPORT_DATA_BLOCK = 0x5,
    /* 0x6 - 0xE reserved */
    NVME_SGL_TYPE_VENDOR_SPECIFIC      = 0xF
};

enum nvme_sgl_descriptor_subtype {
    NVME_SGL_SUBTYPE_ADDRESS           = 0x0,
    NVME_SGL_SUBTYPE_OFFSET            = 0x1,
};

typedef struct NVME_PACKED nvme_sgl_desc {
    uint64_t    address;
    uint32_t    length;
    uint8_t     resv[3];
    uint8_t     subtype : 4;
    uint8_t     type : 4;
} nvme_sgl_desc_t;

typedef struct NVME_PACKED nvme_keyed_sgl_desc {
    uint64_t    address;
    uint64_t    length :24;
    uint64_t    key :32;
    uint64_t    subtype : 4;
    uint64_t    type : 4;
} nvme_keyed_sgl_desc_t;

union nvme_data_ptr {
        struct {
                uint64_t prp1;
                uint64_t prp2;
        };
        nvme_sgl_desc_t    sgl;
        nvme_keyed_sgl_desc_t ksgl;
};

typedef struct NVME_PACKED nvme_psd {
    uint16_t    mp;
    uint16_t    reserved;
    uint32_t    enlat;
    uint32_t    exlat;
    uint8_t     rrt;
    uint8_t     rrl;
    uint8_t     rwt;
    uint8_t     rwl;
    uint8_t     resv[16];
} nvme_psd_t;

typedef struct NVME_PACKED nvme_ctrl_id {
    uint16_t    vid;
    uint16_t    ssvid;
    uint8_t     sn[20];
    uint8_t     mn[40];
    uint8_t     fr[8];
    uint8_t     rab;
    uint8_t     ieee[3];
    uint8_t     cmic;
    uint8_t     mdts;
    uint16_t    cntlid;
    uint32_t    ver;
    uint32_t    rtd3r;
    uint32_t    rtd3e;
    uint32_t    oaes;
    uint8_t     rsvd96[160];
    uint16_t    oacs;
    uint8_t     acl;
    uint8_t     aerl;
    uint8_t     frmw;
    uint8_t     lpa;
    uint8_t     elpe;
    uint8_t     npss;
    uint8_t     avscc;
    uint8_t     apsta;
    uint16_t    wctemp;
    uint16_t    cctemp;
    uint16_t    mtfa;
    uint32_t    hmpre;
    uint32_t    hmmin;
    uint64_t    tnvmcap[2];
    uint64_t    unvmcap[2];
    uint32_t    rpmbs;
    uint16_t    edstt;
    uint8_t     dsto;
    uint8_t     fwug;
    uint16_t    kas;
    uint8_t     rsvd511[190];
    uint8_t     sqes;
    uint8_t     cqes;
    uint16_t    rsvd515;
    uint32_t    nn;
    uint16_t    oncs;
    uint16_t    fuses;
    uint8_t     fna;
    uint8_t     vwc;
    uint16_t    awun;
    uint16_t    awupf;
    uint8_t     nvscc;
    uint8_t     rsvd704;
    uint16_t    acwu;
    uint8_t     rsvd706[2];
    uint32_t    sgls;
    uint8_t     rsvd712[228];
    uint8_t     subnqn[256];
    uint8_t     rsvd2047[1024];
    nvme_psd_t  psd[32];
    uint8_t     vs[1024];
} nvme_ctrl_id_t;

#define NVME_CTRL_SQES_MIN(sqes) ((sqes) & 0xf)
#define NVME_CTRL_SQES_MAX(sqes) (((sqes) >> 4) & 0xf)
#define NVME_CTRL_CQES_MIN(cqes) ((cqes) & 0xf)
#define NVME_CTRL_CQES_MAX(cqes) (((cqes) >> 4) & 0xf)
#define NVME_CTRL_MDTS_DEFAULT   1
#define NVME_CRTL_MDTS_MAX       6

#define NVME_CQE_STATUS_IS_ERROR(status) (le16_to_cpu(status) >> 1)

typedef struct NVME_PACKED nvme_cqe {
    uint32_t    result;
    uint32_t    rsvd;
    uint16_t    sq_head;
    uint16_t    sq_id;
    uint16_t    cid;
    uint16_t    status;
} nvme_cqe_t;

typedef struct NVME_PACKED nvme_cmd {
    uint8_t     opcode;
    uint8_t     fuse : 2; /* fused operation */
    uint8_t     rsvd1 : 4;
    uint8_t     psdt : 2;
    uint16_t    cid;
    uint32_t    nsid;
    uint64_t    res1;
    uint64_t    mptr;
    union nvme_data_ptr dptr;
    uint32_t    cdw10;
    uint32_t    cdw11;
    uint32_t    cdw12;
    uint32_t    cdw13;
    uint32_t    cdw14;
    uint32_t    cdw15;
} nvme_cmd_t;

typedef struct NVME_PACKED nvme_cmd_abort {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint8_t     rsvd4[36];
    uint16_t    abort_sqid;
    uint16_t    abort_cid;
    uint32_t    rsvd40[5];
} nvme_cmd_abort_t;

enum NvmeAbortResult {
    NVME_ABORT_SUCCEEDED = 0x00,
    NVME_ABORT_FAILED    = 0x01
};

typedef struct NVME_PACKED nvme_cmd_get_feature {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint32_t    rsvd24[5];
    uint64_t    prp1;
    uint64_t    rsvd40;
    uint8_t     ftr_id;
    uint8_t     ftr_val;
    uint8_t     rsvd44[2];
    uint32_t    dw11;
    uint32_t    dw12;
    uint32_t    dw13;
    uint32_t    dw14;
    uint32_t    dw15;
} nvme_cmd_get_feature_t;

typedef struct NVME_PACKED nvme_cmd_set_feature {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint32_t    rsvd24[5];
    uint64_t    prp1;
    uint64_t    rsvd40;
    uint8_t     ftr_id;
    uint16_t    rsvd43;
    uint8_t     save;
    uint32_t    dw11;
    uint32_t    dw12;
    uint32_t    dw13;
    uint32_t    dw14;
    uint32_t    dw15;
} nvme_cmd_set_feature_t;

enum NvmeFeatureIds {
    NVME_ARBITRATION                        = 0x1,
    NVME_POWER_MANAGEMENT                   = 0x2,
    NVME_LBA_RANGE_TYPE                     = 0x3,
    NVME_TEMPERATURE_THRESHOLD              = 0x4,
    NVME_ERROR_RECOVERY                     = 0x5,
    NVME_VOLATILE_WRITE_CACHE               = 0x6,
    NVME_NUMBER_OF_QUEUES                   = 0x7,
    NVME_INTERRUPT_COALESCING               = 0x8,
    NVME_INTERRUPT_VECTOR_CONF              = 0x9,
    NVME_WRITE_ATOMICITY                    = 0xa,
    NVME_ASYNCHRONOUS_EVENT_CONF            = 0xb,
    NVME_AUTONOMOUS_POWER_STATE_TRANSIT     = 0xc,
    NVME_HOST_MEM_BUFFER                    = 0xd,
    NVME_TIMESTAMP                          = 0xe,
    NVME_KEEP_ALIVE_TIMER                   = 0xf,
    NVME_HOST_CONTROLLED_THERMAL_MANAGEMENT = 0x10,
    NVME_SOFTWARE_PROGRESS_MARKER           = 0x80,
    NVME_HOST_IDENTIFIER                    = 0x81
};

enum NvmeFeatureValue {
    NVME_FEATURE_CURRENT = 0x0,
    NVME_FEATURE_DEFAULT = 0x1,
    NVME_FEATURE_SAVED   = 0x2,
    NVME_FEATURE_CAPS    = 0x3,
    NVME_FEATURE_MASK    = 0x7
};

enum NvmeFeatureCaps {
    NVME_FEATURE_CHANGEABLE = (1 << 2)
};

#define NVME_FTR_PM_PSTATE(dw11)    ((dw11) & 0xF)

#define NVME_FTR_AGGR_TIME_SHIFT     8
#define NVME_FTR_AGGR_THRESH(dw11)  ((dw11) & 0xFF)
#define NVME_FTR_AGGR_TIME(dw11)    (((dw11) >> NVME_FTR_AGGR_TIME_SHIFT) & 0xFF)

enum NvmeAdminCommands {
    NVME_ADM_CMD_DELETE_SQ      = 0x00,
    NVME_ADM_CMD_CREATE_SQ      = 0x01,
    NVME_ADM_CMD_GET_LOG_PAGE   = 0x02,
    NVME_ADM_CMD_DELETE_CQ      = 0x04,
    NVME_ADM_CMD_CREATE_CQ      = 0x05,
    NVME_ADM_CMD_IDENTIFY       = 0x06,
    NVME_ADM_CMD_ABORT          = 0x08,
    NVME_ADM_CMD_SET_FEATURES   = 0x09,
    NVME_ADM_CMD_GET_FEATURES   = 0x0a,
    NVME_ADM_CMD_ASYNC_EV_REQ   = 0x0c,
    NVME_ADM_CMD_ACTIVATE_FW    = 0x10,
    NVME_ADM_CMD_DOWNLOAD_FW    = 0x11,
    NVME_ADM_CMD_KEEP_ALIVE     = 0x18,
    NVME_ADM_CMD_FORMAT_NVM     = 0x80,
    NVME_ADM_CMD_SECURITY_SEND  = 0x81,
    NVME_ADM_CMD_SECURITY_RECV  = 0x82,
};

enum NvmeIoCommands {
    NVME_CMD_FLUSH              = 0x00,
    NVME_CMD_WRITE              = 0x01,
    NVME_CMD_READ               = 0x02,
    NVME_CMD_WRITE_UNCOR        = 0x04,
    NVME_CMD_COMPARE            = 0x05,
    NVME_CMD_WRITE_ZEROS        = 0x08,
    NVME_CMD_DSM                = 0x09,
};

enum NvmeStatusCodes {
    NVME_SUCCESS                = 0x0000,
    NVME_INVALID_OPCODE         = 0x0001,
    NVME_INVALID_FIELD          = 0x0002,
    NVME_CID_CONFLICT           = 0x0003,
    NVME_DATA_TRAS_ERROR        = 0x0004,
    NVME_POWER_LOSS_ABORT       = 0x0005,
    NVME_INTERNAL_DEV_ERROR     = 0x0006,
    NVME_CMD_ABORT_REQ          = 0x0007,
    NVME_CMD_ABORT_SQ_DEL       = 0x0008,
    NVME_CMD_ABORT_FAILED_FUSE  = 0x0009,
    NVME_CMD_ABORT_MISSING_FUSE = 0x000a,
    NVME_INVALID_NSID           = 0x000b,
    NVME_CMD_SEQ_ERROR          = 0x000c,
    NVME_LBA_RANGE              = 0x0080,
    NVME_CAP_EXCEEDED           = 0x0081,
    NVME_NS_NOT_READY           = 0x0082,
    NVME_NS_RESV_CONFLICT       = 0x0083,
    NVME_INVALID_CQID           = 0x0100,
    NVME_INVALID_QID            = 0x0101,
    NVME_MAX_QSIZE_EXCEEDED     = 0x0102,
    NVME_ACL_EXCEEDED           = 0x0103,
    NVME_RESERVED               = 0x0104,
    NVME_AER_LIMIT_EXCEEDED     = 0x0105,
    NVME_INVALID_FW_SLOT        = 0x0106,
    NVME_INVALID_FW_IMAGE       = 0x0107,
    NVME_INVALID_IRQ_VECTOR     = 0x0108,
    NVME_INVALID_LOG_ID         = 0x0109,
    NVME_INVALID_FORMAT         = 0x010a,
    NVME_FW_REQ_RESET           = 0x010b,
    NVME_INVALID_QUEUE_DEL      = 0x010c,
    NVME_FID_NOT_SAVEABLE       = 0x010d,
    NVME_FID_NOT_NSID_SPEC      = 0x010f,
    NVME_FW_REQ_SUSYSTEM_RESET  = 0x0110,
    NVME_CONFLICTING_ATTRS      = 0x0180,
    NVME_INVALID_PROT_INFO      = 0x0181,
    NVME_WRITE_TO_RO            = 0x0182,
    NVME_WRITE_FAULT            = 0x0280,
    NVME_UNRECOVERED_READ       = 0x0281,
    NVME_E2E_GUARD_ERROR        = 0x0282,
    NVME_E2E_APP_ERROR          = 0x0283,
    NVME_E2E_REF_ERROR          = 0x0284,
    NVME_CMP_FAILURE            = 0x0285,
    NVME_ACCESS_DENIED          = 0x0286,
    NVME_MORE                   = 0x2000,
    NVME_DNR                    = 0x4000,
    NVME_NO_COMPLETE            = 0xffff,
};

enum NvmeAsyncEventConfig {
    NVME_AVAILABLE_SPARE_ERR     = 1 << 0,
    NVME_TEMPERATURE_ERR         = 1 << 1,
    NVME_DEV_RELIABILITY_ERR     = 1 << 2,
    NVME_READ_ONLY_ERR           = 1 << 3,
    NVME_VOLATILE_MEM_BACKUP_ERR = 1 << 4,
    NVME_NAMESPACE_ATTR_NOTICE   = 1 << 8,
    NVME_FW_ACTIVATION_NOTICE    = 1 << 9,
    NVME_TELEMETRY_LOG_NOTICE    = 1 << 10,
};

enum NvmeLogPage {
    NVME_LOG_ERROR                     = 0x01,
    NVME_LOG_HEALTH_INFORMATION        = 0x02,
    NVME_LOG_FIRMWARE_SLOT             = 0x03,
    NVME_LOG_CHANGED_NS_LIST           = 0x04,
    NVME_LOG_COMMAND_EFFECTS_LOG       = 0x05,
    NVME_LOG_DISCOVERY                 = 0x70,
    NVME_LOG_RESERVATION_NOTIFICATION  = 0x80,
};

enum NvmeAsyncEventType {
    NVME_ASYNC_EVENT_TYPE_ERROR        = 0x0,
    NVME_ASYNC_EVENT_TYPE_SMART        = 0x1,
    NVME_ASYNC_EVENT_TYPE_NOTICE       = 0x2,
    NVME_ASYNC_EVENT_TYPE_IO           = 0x6,
    NVME_ASYNC_EVENT_TYPE_VENDOR       = 0x7,
};

enum NvmeAsyncEventInfoNotice {
    NVME_ASYNC_EVENT_NS_ATTR_CHANGED           = 0x0,
    NVME_ASYNC_EVENT_FW_ACTIVATION_START       = 0x1,
    NVME_ASYNC_EVENT_TELEMETRY_LOG_CHANGED     = 0x2,
};

enum NvmeIdentifyCns {
    NVME_IDENTIFY_CNS_NAMESPACE             = 0x00,
    NVME_IDENTIFY_CNS_CTRL                  = 0x01,
    NVME_IDENTIFY_CNS_ACTIVE_NS_LIST        = 0x02,
    NVME_IDENTIFY_CNS_NS_ID_DESCRIPTOR_LIST = 0x03,
};

enum NvmeIdentifyCtrlOncs {
    NVME_ONCS_COMPARE       = 1 << 0,
    NVME_ONCS_WRITE_UNCORR  = 1 << 1,
    NVME_ONCS_DSM           = 1 << 2,
    NVME_ONCS_WRITE_ZEROS   = 1 << 3,
    NVME_ONCS_FEATURES      = 1 << 4,
    NVME_ONCS_RESERVATIONS  = 1 << 5,
    NVME_ONCS_TIMESTAMP     = 1 << 6,
};

typedef struct NVME_PACKED nvme_fw_slot_info_log {
    uint8_t     afi;
    uint8_t     reserved1[7];
    uint8_t     frs1[8];
    uint8_t     frs2[8];
    uint8_t     frs3[8];
    uint8_t     frs4[8];
    uint8_t     frs5[8];
    uint8_t     frs6[8];
    uint8_t     frs7[8];
    uint8_t     reserved2[448];
} nvme_fw_slot_info_log_t;

typedef struct NVME_PACKED nvme_error_log {
    uint64_t    error_count;
    uint16_t    sqid;
    uint16_t    cid;
    uint16_t    status_field;
    uint16_t    param_error_location;
    uint64_t    lba;
    uint32_t    nsid;
    uint8_t     vs;
    uint8_t     resv[35];
} nvme_error_log_t;

typedef struct NVME_PACKED nvme_smart_log {
    uint8_t     critical_warning;
    uint16_t    temperature;
    uint8_t     available_spare;
    uint8_t     available_spare_threshold;
    uint8_t     percentage_used;
    uint8_t     reserved1[26];
    uint64_t    data_units_read[2];
    uint64_t    data_units_written[2];
    uint64_t    host_read_commands[2];
    uint64_t    host_write_commands[2];
    uint64_t    controller_busy_time[2];
    uint64_t    power_cycles[2];
    uint64_t    power_on_hours[2];
    uint64_t    unsafe_shutdowns[2];
    uint64_t    media_errors[2];
    uint64_t    number_of_error_log_entries[2];
    uint8_t     reserved2[320];
} nvme_smart_log_t;

typedef struct NVME_PACKED nvme_changed_nslist_log {
    uint32_t     ns_list[1024];
} nvme_changed_nslist_log_t;

enum NvmeLogIdentifier {
    NVME_LOG_ERROR_INFO             = 0x01,
    NVME_LOG_SMART_INFO             = 0x02,
    NVME_LOG_FW_SLOT_INFO           = 0x03,
    NVME_LOG_CHANGED_NS_LIST_INFO   = 0x04,
};

typedef struct NVME_PACKED nvme_cmd_identify {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint32_t    nsid;
    uint64_t    rsvd2[2];
    uint64_t    prp1;
    uint64_t    prp2;
    uint8_t     cns;
    uint8_t     rsvd3;
    uint16_t    ctrlid;
    uint32_t    rsvd11[5];
} nvme_cmd_identify_t;

enum NvmeNamespaceIdentifierType {
    NVME_NS_IDENTIFIER_EUI64  = 0x1,
    NVME_NS_IDENTIFIER_NGUID  = 0x2,
    NVME_NS_IDENTIFIER_UUID   = 0x3,
    NVME_NS_IDENTIFIER_LAST
};

#define NVME_NS_ID_TYPES_NUM (NVME_NS_IDENTIFIER_LAST - 1)

typedef struct NVME_PACKED nvme_eui64 {
    uint8_t raw[8];
} nvme_eui64_t;

typedef struct NVME_PACKED nvme_nguid {
    uint8_t raw[16];
} nvme_ngiud_t;

typedef struct NVME_PACKED nvme_uuid {
    uint32_t      time_low;
    uint16_t      time_mid;
    uint16_t      time_hi_and_version;
    uint16_t      clk_seq; /* hi_res + low */
    struct {
        uint32_t  byte5;
        uint16_t  byte1;
    } node;
} nvme_uuid_t;

typedef union NVME_PACKED nvme_id_ns_global {
        nvme_eui64_t  eui64;
        nvme_ngiud_t  nguid;
        nvme_uuid_t   uuid;
} nvme_id_ns_global_t;

typedef struct NVME_PACKED nvme_id_ns_descriptor {
    uint8_t              nidt;
    uint8_t              nidl;
    uint8_t              rsvd[2];
    nvme_id_ns_global_t  nid;
} nvme_id_ns_descriptor_t;

typedef struct NVME_PACKED nvme_cmd_create_cq {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint32_t    rsvd1[5];
    uint64_t    prp1;
    uint64_t    rsvd8;
    uint16_t    cqid;
    uint16_t    qsize;
    uint16_t    cq_flags;
    uint16_t    irq_vector;
    uint32_t    rsvd12[4];
} nvme_cmd_create_cq_t;

typedef struct NVME_PACKED nvme_cmd_get_page_log {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint32_t    nsid;
    uint32_t    rsvd24[4];
    uint64_t    prp1;
    uint64_t    rsvd40;
    uint8_t     page_id;
    uint8_t     lsp;
    uint16_t    num_dwords;
    uint32_t    rsvd60[4];
} nvme_cmd_get_page_log_t;

#define NVME_CQ_FLAGS_PC(cq_flags)  (cq_flags & 0x1)
#define NVME_CQ_FLAGS_IEN(cq_flags) ((cq_flags >> 1) & 0x1)

typedef struct NVME_PACKED nvme_cmd_create_sq {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint32_t    rsvd1[5];
    uint64_t    prp1;
    uint64_t    rsvd8;
    uint16_t    sqid;
    uint16_t    qsize;
    uint16_t    sq_flags;
    uint16_t    cqid;
    uint32_t    rsvd12[4];
} nvme_cmd_create_sq_t;

#define NVME_SQ_FLAGS_PC(sq_flags)      (sq_flags & 0x1)
#define NVME_SQ_FLAGS_QPRIO(sq_flags)   ((sq_flags >> 1) & 0x3)

enum NvmeQueueFlags {
    NVME_Q_PC           = 1,
    NVME_Q_PRIO_URGENT  = 0,
    NVME_Q_PRIO_HIGH    = 1,
    NVME_Q_PRIO_NORMAL  = 2,
    NVME_Q_PRIO_LOW     = 3,
};

typedef struct NVME_PACKED nvme_cmd_delete_q {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint32_t    rsvd1[9];
    uint16_t    qid;
    uint16_t    rsvd10;
    uint32_t    rsvd11[5];
} nvme_cmd_delete_q_t;

typedef struct NVME_PACKED nvme_cmd_rw {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint32_t    nsid;
    uint64_t    rsvd2;
    uint64_t    mptr;
    union nvme_data_ptr dptr;
    uint64_t    slba;
    uint16_t    nlb;
    uint16_t    control;
    uint32_t    dsmgmt;
    uint32_t    reftag;
    uint16_t    apptag;
    uint16_t    appmask;
} nvme_cmd_rw_t;

enum {
    NVME_DSM_IDR = 1 << 0,
    NVME_DSM_IDW = 1 << 1,
    NVME_DSM_AD  = 1 << 2,
};

typedef struct NVME_PACKED nvme_dsm_range {
    uint8_t  access;
    uint8_t  flags;
    uint8_t  rsvd3;
    uint8_t  cas;
    uint32_t nlb;
    uint64_t slba;
} nvme_dsm_range_t;

typedef struct NVME_PACKED nvme_cmd_dsm {
    uint8_t     opcode;
    uint8_t     flags;
    uint16_t    cid;
    uint32_t    nsid;
    uint8_t     rsvd8[16];
    union nvme_data_ptr dptr;
    uint8_t     nr;
    uint8_t     rsvd25[3];
    uint8_t     attributes;
    uint8_t     rsvd41[19];
} nvme_cmd_dsm_t;

typedef struct NVME_PACKED nvme_aen_completion  {
    uint32_t ev_type : 3;
    uint32_t reserved1 : 5;
    uint32_t ev_info : 8;
    uint32_t log_page_identifier : 8;
    uint32_t reserved2 : 8;
} nvme_aen_completion_t;

#endif
