#ifndef _NVME_EMU_H
#define _NVME_EMU_H

#include <stdint.h>
#include <string.h>
#include <stdbool.h>
#include <errno.h>

#include "config.h"
#include "nvme_emu_log.h"
#include "nvme_regs.h"
#include "json/json.h"
#include "queue.h"
#include "utils.h"

#define NVME_DISK_SIZE_MB_VAR           "NVME_DISK_SIZE_MB"
#define NVME_LBA_SIZE_VAR               "NVME_LBA_SIZE"
#define NVME_METADATA_SIZE_VAR          "NVME_METADATA_SIZE"

#define NVME_DEFAULT_POLLING_TIME_INTERVAL_MSEC  10
#define NVME_POLLING_TIME_INTERVAL_MSEC          "NVME_POLLING_TIME_INTERVAL"


enum {
    NVMX_OK  =  0,
    NVMX_ERR = -1,
    NVMX_ERR_NOTSUPPORTED = -2,
    NVMX_ERR_NOTREADY = -3
};

#define NVME_EMU_DB_STRIDE       8

typedef struct nvme_emulator        nvme_emulator_t;
/* Declaration from driver and ep */
typedef struct nvme_emu_ep nvme_emu_ep_t;
typedef struct nvme_emu_driver nvme_emu_driver_t;
typedef struct nvme_emu_driver_queue nvme_emu_driver_queue_t;
typedef struct nvme_queue_ops       nvme_queue_ops_t;
typedef struct nvme_emu_ev_data     nvme_emu_ev_data_t;
typedef struct nvme_emu_dev_ns      nvme_emu_dev_ns_t;

/* Called when host has written to the bar config space. 
 * Where addr >= 0x0 and addr < 0x50 */
typedef void (*nvme_emu_mmio_write_cb_t)(void *ctrl);

/* Called when host has written to the door bell register.
 * db_addr is an offset from the beginning of the bar
 */
typedef void (*nvme_emu_db_cb_t)(nvme_queue_ops_t *ops, uint32_t db_addr, uint32_t val);

enum nvme_emu_queue_type {
    NVME_EMU_QUEUE_DB,
    NVME_EMU_QUEUE_SQE,  /* queue receives SQ entries */
    NVME_EMU_QUEUE_NVME2NVMF_LOCAL,
    NVME_EMU_QUEUE_NVME2NVMF_REMOTE
};

enum namespace_mgmt_type {
    NAMESPACE_MGMT_TYPE_STATIC,
    NAMESPACE_MGMT_TYPE_DYNAMIC,
};

typedef int attach_namespace_fn_t(void *ctrl, uint32_t nsid, uint64_t size_b,
                                  uint8_t block_order, uint16_t backend_metadata,
                                  nvme_id_ns_descriptor_t id_descs[]);
typedef int detach_namespace_fn_t(void *ctrl, uint32_t nsid);

struct nvme_emu_dev_ns {
    void *ns;//Low level namespace
    nvme_emulator_t *dev;
};

void nvme_emu_dev_remove_namespace(nvme_emu_dev_ns_t *ns);
nvme_emu_dev_ns_t *nvme_emu_dev_add_namespace(nvme_emulator_t *dev,
                                uint32_t src_nsid, uint32_t dst_nsid,
                                uint16_t md_size, uint8_t lba_size);

struct nvme_queue_ops {
    int  q_type; /* queue type */
    void *q_impl;/* actual queue implementation */
    int  (*process_sq)(nvme_queue_ops_t *ops, uint32_t db_addr, void *cmd, uint32_t len);
    int  (*process_cq)(nvme_queue_ops_t *ops, uint32_t db_addr, void *cmd, uint32_t len);

    int              qid;
    nvme_emulator_t *dev;

    /* not currently used */
    int  (*nvme_queue_init)();
    void (*nvme_queue_destroy)();
};

typedef struct nvme_emu_ops {
    /* open NVME emulation device */
    nvme_emulator_t * (*open)(const char *dev_name, nvme_config_t *config,
                              nvme_emu_mmio_write_cb_t bar_write_ev_handler,
                              void *dev_ctx, uint32_t *num_queues);
    /* close NVME emulation device */
    void (*close)(nvme_emulator_t *dev);

    /* start NVMe emulation device */
    int (*start)(nvme_emulator_t *dev);

    /* stop NVMe emulation device */
    int (*stop)(nvme_emulator_t *dev);

    /* check if NVMe emulation device started */
    bool (*is_started)(nvme_emulator_t *dev);

    /* read from host bar space */
    int (*mmio_read)(nvme_emulator_t *dev, void *buf, uint32_t bar_addr,
                     unsigned len);

    /* write to host bar space */
    int (*mmio_write)(nvme_emulator_t *dev, void *buf, uint32_t bar_addr,
                      unsigned len);

    /* create/destroy SQ/CQ
     * On succes the ops structure is filled
     */
    int (*create_sq)(nvme_emulator_t *dev, int qid, uint64_t base_addr,
                     uint16_t size, nvme_queue_ops_t *cq_ops,
                     nvme_queue_ops_t *ops, uint8_t lba_size,
                     uint16_t md_size);

    int (*create_cq)(nvme_emulator_t *dev, int qid, uint64_t base_addr,
                     uint16_t size, int vector, int cq_period,
                     int cq_max_count, nvme_queue_ops_t *ops);

    /* remote sq or cq */
    int (*delete_sq)(nvme_queue_ops_t *ops, int sqid);
    int (*delete_cq)(nvme_queue_ops_t *ops, int cqid);

    /* dma read/write */
    void (*dma_read)(nvme_queue_ops_t *ops, void *dst_buf, uint64_t srcaddr,
                     size_t len, uint32_t rmkey);
    void (*dma_write)(nvme_queue_ops_t *ops, void *src_buf, uint64_t dstaddr,
                      size_t len, uint32_t rmkey);
    /* send to host memory */
    void (*send)(nvme_queue_ops_t *ops, void *src_buf, size_t len);

    /* generate interrupt on host */
    void (*interrupt)(nvme_queue_ops_t *ops);

    /* progress nvme emulation device */
    int (*progress)(nvme_emulator_t *dev);

    nvme_emu_dev_ns_t * (*add_namespace)(nvme_emulator_t *dev, uint32_t src_nsid,
                                         uint32_t dst_nsid, uint16_t md_size,
                                         uint8_t lba_size);
    void (*remove_namespace)(nvme_emu_dev_ns_t *ns);
} nvme_emu_ops_t;

/* add/remove a queue layer implementation 
 * if is_adm is true the layer applies to the admin qps
 * TODO: extend to filter/priority
 */
int nvme_emu_queue_type_add(nvme_emulator_t *dev, int queue_type, bool is_adm,
                            nvme_queue_ops_t *ops);
void nvme_emu_queue_type_del(nvme_emulator_t *dev, int queue_type, bool is_adm);

/* create/destroy SQ/CQ
 * On succes the ops structure is filled 
 */
int nvme_emu_create_sq(nvme_emulator_t *dev, int qid, uint64_t base_addr,
                       uint16_t size, nvme_queue_ops_t *cq_ops,
                       nvme_queue_ops_t *ops, uint8_t lba_size,
                       uint16_t md_size);
int nvme_emu_create_cq(nvme_emulator_t *dev, int qid, uint64_t base_addr,
                       uint16_t size, int vector, int cq_period,
                       int cq_max_count, nvme_queue_ops_t *ops);
/* remote sq or cq */
int nvme_emu_delete_sq(nvme_queue_ops_t *ops, int sqid);
int nvme_emu_delete_cq(nvme_queue_ops_t *ops, int cqid);

static inline int nvme_emu_process_sq(nvme_queue_ops_t *ops, uint32_t db_addr, void *cmd, uint32_t len)
{
    return ops->process_sq(ops, db_addr, cmd, len);
}

static inline int nvme_emu_process_cq(nvme_queue_ops_t *ops, uint32_t db_addr, void *cmd, uint32_t len)
{
    return ops->process_cq(ops, db_addr, cmd, len);
}

typedef struct nvme_query_attr {
    /* Device Capabilites */
    uint32_t    max_namespaces;         /**< max number of namespaces support per SQ */
    uint16_t    max_reg_size;       /**< max size of the bar register area */
    uint32_t    max_emulated_cq_num;    /**< max number of CQs that can be emulated */
    uint32_t    max_emulated_sq_num;    /**< max number of SQs that can be emulated */
    uint32_t    max_emulated_pfs;    /**< max number of NVMe function that can be emulated */
} nvme_query_attr_t;

struct nvme_emulator {
    /* Mellanox Emulation SDK */
    void                        *ctrl;   /* points to controller */
    nvme_emu_mmio_write_cb_t     bar_write_host_cb;
    nvme_bar_instance_t          bar;
    nvme_queue_ops_t             *admin_q_ops;
    nvme_queue_ops_t             *io_q_ops;
    nvme_emu_ops_t               *emu_ops;
    nvme_query_attr_t            dev_caps;
    uint16_t                     vid;

    nvme_config_t                *config;
    char                         if_name[IFACE_MAX_LEN];
};

/* NVME emulation functions */

/* Opening this device will use simple in memory
 * emulation of the nvme controller.
 */
#define NVME_EMU_TEST_DEV "nvme_test"

/**
 * Setup NVME emulator
 * The function will do following things:
 * - check that the device exists
 * - check device attributes: (dev_emu and cmd_on_behalf)
 * - initialize NVME emulation (init hca on behalf of host)
 * - create QP that to pass doorbells and to read/write from
 *   the host memory
 *
 * return:
 *
 * Emulator object on success, Otherwise NULL.
 */
nvme_emulator_t *nvme_emu_dev_open(const char *iface_name,
                                   nvme_config_t *config,
                                   nvme_emu_mmio_write_cb_t bar_write_ev_handler,
                                   void *dev_ctx,
                                   uint32_t *num_queues);

/* close NVME emulation device */
void nvme_emu_dev_close(nvme_emulator_t *dev);

/* start NVMe emulation device */
int nvme_emu_dev_start(nvme_emulator_t *dev);

/* stop NVMe emulation device */
int nvme_emu_dev_stop(nvme_emulator_t *dev);

/* check if NVMe emulation device started */
bool nvme_emu_dev_is_started(nvme_emulator_t *dev);

/* read from host bar space */
int nvme_emu_mmio_read(nvme_emulator_t *dev, void *buf, uint32_t bar_addr,
		               unsigned len);

/* write to host bar space */ 
int nvme_emu_mmio_write(nvme_emulator_t *dev, void *buf, uint32_t bar_addr,
                        unsigned len);


/* do dma transfers to/from hst memory */
/* read from the host memory */
void nvme_emu_dma_read(nvme_queue_ops_t *ops, void *dst_buf, uint64_t srcaddr,
                       size_t len, uint32_t rmkey);
/* write to the host memory */
void nvme_emu_dma_write(nvme_queue_ops_t *ops, void *src_buf, uint64_t dstaddr,
                        size_t len, uint32_t rmkey);
/* send to host memory */
void nvme_emu_send(nvme_queue_ops_t *ops, void *src_buf, size_t len);

/* generate interrupt on host */
void nvme_emu_interrupt(nvme_queue_ops_t *ops);

/* progress nvme emulation device:
 *  check and dispatch new events (dbs, bar write)
 *  progress outstanding dma transfers
 *
 *  return nummber of event handled
 */
int nvme_emu_progress(nvme_emulator_t *dev);

/**
 * internal function but it is here to avoid linking problems when using from
 * different nvme layer implementations
 */
static inline void nvme_emu_qops_clone(nvme_emulator_t *dev,
                                       nvme_queue_ops_t *ops, int qid)
{
    if (qid)
        memcpy(ops, dev->io_q_ops, sizeof(*ops));
    else
        memcpy(ops, dev->admin_q_ops, sizeof(*ops));

    ops->qid = qid;
    ops->dev = dev;
}

/**
 * Helper function to launch doorbell emulation mode callbacks
 */
static inline void nvme_emu_db_dispatch(nvme_queue_ops_t *ops, uint32_t addr, uint32_t val)
{
    if (addr & ((1 << 2) - 1)) {
        nvmx_error("doorbell write not 32-bit aligned,"
                   " offset=0x%x ignoring", addr);
        return;
    }

    if (((addr - 0x1000) >> 2) & 1) {
        /* Completion queue doorbell write */
        nvme_emu_process_cq(ops, addr, (void *)(uintptr_t)val, sizeof(val));
    } else {
        /* Submission queue doorbell write */
        nvme_emu_process_sq(ops, addr, (void *)(uintptr_t)val, sizeof(val));
    }
}

/**
 * Helper function to set number of queues according to caps
 */
void nvme_emu_set_num_queues(nvme_query_attr_t *dev_caps,
                             nvme_config_t *config,
                             uint32_t *num_queues);

#endif
