#ifndef _NVME_EMU_IO_DRIVER_H
#define _NVME_EMU_IO_DRIVER_H

#include <stdlib.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "nvme.h"
#include "nvme_emu.h"

#include "json/json.h"

 /* 512 MB is default disk size
 * Can be changed via env var NVME_DISK_SIZE
 */
#define NVME_DISK_DEFAULT_SIZE    (512 * 1024 * 1024ULL)
#define NVME_DISK_DEFAULT_LBA      9
#define NVME_DISK_DEFAULT_METADATA 0
#define NVME_DISK_SIZE_VAR        "NVME_IO_DISK_SIZE"

enum nvme_emu_driver_state {
    NVME_EMU_DRIVER_NEW           = 0x0,
    NVME_EMU_DRIVER_ACTIVE        = 0x1,
    NVME_EMU_DRIVER_NON_ACTIVE    = 0x2,
    NVME_EMU_DRIVER_DELETING      = 0x3,
    NVME_EMU_DRIVER_DEAD          = 0x4,
};

/*
 * NVMe Emulation driver types
 */
enum nvme_emu_driver_type {
    /* single backend driver */
    NVME_EMU_SINGLE_EP_DRIVER_TYPE   = 0x1,
    /* HA driver */
    NVME_EMU_MULTIPLE_EP_DRIVER_TYPE = 0x2,
};

enum nvme_emu_driver_queue_type {
    NVME_EMU_DRIVER_SQ   = 0x1,
    NVME_EMU_DRIVER_CQ   = 0x2,
};

struct nvme_json_val;

struct nvme_emu_driver_queue {
    nvme_emu_driver_t *driver;
    enum nvme_emu_driver_queue_type type;
    uint16_t qid;
    uint16_t depth;

    /* valid for NVME_EMU_DRIVER_SQ */
    uint16_t cqid;
};

/*
 * Template structure that each underling layer should supply to
 * the Nvme ctrl for processing data received from the host OS.
 */
struct nvme_emu_driver {
    enum nvme_emu_driver_type     type;
    enum nvme_emu_driver_state    state;
    pthread_mutex_t               state_lock;
    int                           nr_endpoints;
    int                           active_endpoint;
    pthread_mutex_t               ep_lock;
    nvme_emu_ep_t                 *endpoints;
    nvme_emulator_t               *dev;
    enum namespace_mgmt_type      ns_mgmt_type;
    uint32_t                      ver;
    uint8_t                       disk_block_order;
    uint16_t                      disk_md_size;
    uint8_t                       mdts;
    int                           num_queues;
    int                           num_created_io_squeues;
    int                           num_created_io_cqueues;
    nvme_emu_driver_queue_t       *sqs;
    nvme_emu_driver_queue_t       *cqs;
    nvme_config_t                 *config;
    pthread_t                     d_thread;
    sem_t                         sem;

    /* Controller callbacks */
    attach_namespace_fn_t            *attach_namespace_cb;
    detach_namespace_fn_t            *detach_namespace_cb;
};

void nvme_driver_destroy(nvme_emu_driver_t *driver);
nvme_emu_driver_t *nvme_driver_create(nvme_emulator_t *dev,
                                      nvme_config_t *config,
                                      int num_queues,
                                      attach_namespace_fn_t *attach_namespace_fn,
                                      detach_namespace_fn_t *detach_namespace_fn,
                                      int index, uint32_t ver, uint8_t mdts);
int nvme_driver_write_prp(nvme_emu_driver_t *io_driver,
                          nvme_queue_ops_t *ops,
                          nvme_cmd_rw_t *rw,
                          int page_bits);
int nvme_driver_read_prp(nvme_emu_driver_t *io_driver,
                         nvme_queue_ops_t *ops,
                         nvme_cmd_rw_t *rw,
                         int page_bits);
int nvme_driver_write_sgl(nvme_emu_driver_t *io_driver, nvme_queue_ops_t *ops,
                          nvme_cmd_rw_t *rw, int page_bits);
int nvme_driver_read_sgl(nvme_emu_driver_t *io_driver, nvme_queue_ops_t *ops,
                         nvme_cmd_rw_t *rw, int page_bits);


int nvme_driver_create_sq(nvme_emu_driver_t *driver, uint16_t sqid,
                          uint64_t dma_addr, uint16_t size,
                          nvme_queue_ops_t *cq_ops, nvme_queue_ops_t *sq_ops);
int nvme_driver_create_cq(nvme_emu_driver_t *driver, int qid,
                          uint64_t dma_addr,uint16_t size, uint16_t vector,
                          int cq_period, int cq_max_count,
                          nvme_queue_ops_t *cq_ops);
void nvme_driver_delete_sq(nvme_emu_driver_t *driver, nvme_queue_ops_t *sq_ops,
                           int qid);
void nvme_driver_delete_cq(nvme_emu_driver_t *driver, nvme_queue_ops_t *cq_ops,
                           int qid);
void nvme_driver_fail_ep(nvme_emu_driver_t *driver, nvme_emu_ep_t *ep);
void nvme_driver_attach_ep_volume(nvme_emu_driver_t *driver, uint32_t nsid,
                                  uint64_t size_b, uint8_t block_order,
                                  uint16_t backend_metadata,
                                  nvme_id_ns_descriptor_t id_descs[]);
void nvme_driver_detach_ep_volume(nvme_emu_driver_t *driver, uint32_t nsid);

#endif
