#!/bin/bash -eE
# Copyright (c) 2019 Mellanox Technologies. All rights reserved.
# Create one NVMe SF - mlx5_2 and start nvme_snap@mlx5_2

function setup_sf()
{
    cfg_fn=$1
    shift
    if [ ! -e "$cfg_fn" ] ; then
        logger -p local0.err -t nvme_snap "$0: can't find config: $cfg_fn"
        # EX_OSFILE == 72
        return 72
    fi
    unset NVME_SF_MACADDR
    unset NVME_SF_IPADDR
    unset NVME_SF_ECPF_DEV
    # load settings from $cfg_fn
    source $cfg_fn
    if [ -z "$NVME_SF_ECPF_DEV" ] ; then
        export NVME_SF_ECPF_DEV=0000:03:00.0
    fi
    chk=$(mlxconfig -d "$NVME_SF_ECPF_DEV" q NVME_EMULATION_ENABLE |awk '/NVME_EMULATION_ENABLE/ {print $2}')
    if [ "$chk" != "True(1)" ] ; then
        logger -p local0.notice -t nvme_snap "$prog: NVMe emulation is disbled for $NVME_SF_ECPF_DEV"
        echo "NVMe emulation is disbled for $NVME_SF_ECPF_DEV" >&2
        return 69 # EX_UNAVAILABLE
    fi

    p_num=${NVME_SF_ECPF_DEV#*.}
    if [ -z "$NVME_SF_IPADDR" ] ; then
        export NVME_SF_IPADDR=$(printf "192.168.%d.1/30" $[ 101 + $p_num ])
    fi
    if [ -z "$NVME_SF_MACADDR" ] ; then
        # EX_CONFIG == 78
        logger -p local0.err -t nvme_snap "NVME_SF_MACADDR is not set in cfg_fn. Exit"
        return 78
    else
        # specified manually - based on NVME_SF_MACADDR
        NVME_SF_MACADDR=$(echo $NVME_SF_MACADDR | \
                         perl -pne 's/^(.*):([0-9a-f]{2})$/sprintf("%s:%02x", $1, hex($2) + $p_num)/e')
    fi
    export NVME_SF_MACADDR
    create_sf
}

function create_sf()
{
  # parameters (env vars):
  #   DEV  - 0000:03:00.0 (port 1) or 0000:03:00.1 (port 2)
  devpath=/sys/bus/pci/devices/$NVME_SF_ECPF_DEV
  fn="/dev/shm/.nvme_sf_guid-$NVME_SF_ECPF_DEV"
  if [ -e "$fn" ] ; then
    guid=$(cat "$fn")
  else
    guid=$(uuidgen)
    echo $guid > "$fn"
  fi
  if [ ! -e "${devpath}/mdev_supported_types/mlx5_core-local/max_mdevs" ]  ; then
    logger -p local0.err -t nvme_snap "$0: can't configure SF using $NVME_SF_ECPF_DEV - BAR2 is disabled"
    # EX_UNAVAILABLE
    return 69
  fi
  x=$(cat ${devpath}/mdev_supported_types/mlx5_core-local/max_mdevs)
  if [ "$x" -ne 2 ] ;then
    echo 2 > ${devpath}/mdev_supported_types/mlx5_core-local/max_mdevs
  fi
  spath=${devpath}/mdev_supported_types/mlx5_core-local
  if [ ! -e "$spath/devices/$guid" ] ; then
    echo ${guid} > $spath/create
    sleep 1
  else
    logger -p local0.info -t nvme_snap "$0: SF $spath/devices/$guid is already configured"
    return 0
  fi
  echo ${guid} > /sys/bus/mdev/drivers/vfio_mdev/unbind
  p_num=${NVME_SF_ECPF_DEV#*.}
  port="p$p_num"
  p0path=/sys/class/net/$port
  logger -p local0.info -t nvme_snap "$0: ecpf $NVME_SF_ECPF_DEV: NVME_SF_MACADDR set to $NVME_SF_MACADDR"
  
  echo ${NVME_SF_MACADDR} > /sys/bus/mdev/devices/${guid}/devlink-compat-config/mac_addr
  sleep 1
  # find netdev
  netdev=$(cat /sys/bus/mdev/devices/${guid}/devlink-compat-config/netdev)
  cnt=10
  while [ "$netdev" == "pf" -a $cnt -gt 0 ] ; do
     sleep 1
     netdev=$(cat /sys/bus/mdev/devices/${guid}/devlink-compat-config/netdev)
     cnt=$[ cnt - 1 ]
  done
  if [ "$netdev" == "pf" ] ; then
      logger -p local0.err -t nvme_snap "$0: netdev wasn't renamed from 'pf'"
      # EX_DATAERR
      return 69
  fi
  echo ${guid} > /sys/bus/mdev/drivers/mlx5_core/bind
  rdma_dev[$p_num]=$(basename `ls -d /sys/bus/mdev/devices/${guid}/infiniband/* |head -1`)
  logger -p local0.info -t nvme_snap "$0: rdma_dev is ${rdma_dev[$p_num]}"
  dev=$(cat /sys/class/infiniband/${rdma_dev[$p_num]}/ports/1/gid_attrs/ndevs/0)
  logger -p local0.notice -t nvme_snap "$0: NVMe SF netdev '$netdev' is $dev"
  if [ -e /etc/sysconfig/network-scripts/ifcfg-$dev ] ; then
    /sbin/ifup $dev || true
  else
    # bring up network interface using OOB ip-address from $cfg_fn
    if ! grep -qE '/[1-9]' <<< "$NVME_SF_IPADDR" ; then
       NVME_SF_IPADDR="$NVME_SF_IPADDR/30"
    fi
    /sbin/ifconfig $dev $NVME_SF_IPADDR up
  fi 
  #logger -p local0.notice -t nvme_snap "$prog: Enable nvme_snap@${rdma_dev[$p_num]}.service"
  #/usr/bin/systemctl --no-block start nvme_snap@${rdma_dev[$p_num]}.service
}

# Create NVMe SF(s)
for cfg_fn in /etc/nvme_snap/sf*.conf ; do
  setup_sf $cfg_fn
done
